from time import sleep
from rich.panel import Panel
from rich.progress import Progress


JOBS = [100, 150, 25]

progress = Progress(auto_refresh=False)
master_task = progress.add_task("overall", total=sum(JOBS))
jobs_task = progress.add_task("jobs")

jobname = "finetune"
progress.console.print(
    Panel(
        f"[bold blue]Scheduling job {jobname}.",
        padding=1,
    )
)
jobs = [
    "Establish connection with booster",
    "Send source code of job",
    "Submit on job on Slurm",
]
with progress:
    for job_no, job in enumerate(JOBS):
        progress.log(f"Starting job #{job_no} {jobs[job_no]}")
        sleep(0.2)
        progress.reset(jobs_task, total=job, description=f"job [bold yellow]#{job_no}")
        progress.start_task(jobs_task)
        for wait in progress.track(range(job), task_id=jobs_task):
            sleep(0.01)
        progress.advance(master_task, job)
        progress.log(f"Job #{job_no} is complete")
    progress.log(
        Panel(":sparkle: All done! :sparkle:", border_style="green", padding=1)
    )



from pathlib import Path


def slurmpilot():
    res = Path("~/slurmpilot").expanduser()
    res.mkdir(exist_ok=True)
    return res


class JobPathLogic:
    # Could also pass a JobCreationInfo, we dont do it here as it contains other stuff related to ressources etc not
    # relevant here
    def __init__(
        self,
        jobname: str,
        entrypoint: str | None = None,
        src_dir_name: str | None = None,
        root_path: str | None = None,
    ):
        f"""
        performs bookeeping path logic, valid both locally and remotely.
        :param root_path: path where slurmpilot is available
        :param jobname: name of the job, can nest subfolders, for instance "group_experiment_1/experiment2" is allowed 
        and the folder of the job will then be "~/slurmpilot/jobs/group_experiment_1/experiment2".
        :param src_dir_name: directory where the entrypoint is found
        :param entrypoint: path to the entrypoint starting at `src_dir_name`, e.g. f"{src_dir_name}/{entrypoint}" 
        should exists
        """
        self.root_path = root_path if root_path else str(slurmpilot())
        self.jobname = jobname
        self.entrypoint = entrypoint
        self.src_dir_name = src_dir_name

    @classmethod
    def from_jobname(cls, jobname: str, root_path: Path = None):
        return cls(
            jobname=jobname,
            root_path=root_path,
        )

    def job_path(self) -> Path:
        return self.resolve_path()

    def resolve_path(self, path: Path | str | None = None) -> Path:
        if path is not None:
            return Path(self.root_path) / "jobs" / self.jobname / path
        else:
            return Path(self.root_path) / "jobs" / self.jobname

    def slurmpilot_path(self) -> Path:
        return Path(self.root_path)

    def metadata_path(self) -> Path:
        return self.resolve_path("metadata.json")

    def jobid_path(self):
        return self.resolve_path("jobid.json")

    def log_path(self) -> Path:
        return self.resolve_path("logs")

    def stderr_path(self: str) -> Path:
        return self.log_path() / "stderr"

    def stdout_path(self) -> Path:
        return self.log_path() / "stdout"

    def src_path(self) -> Path:
        return self.resolve_path(self.src_dir_name)

    def slurm_entrypoint_path(self) -> Path:
        return self.resolve_path() / "slurm_script.sh"

    def entrypoint_path_from_cwd(self):
        # TODO ugly name, can we avoid it?
        return Path(self.src_path().name) / self.entrypoint



from time import sleep

from rich.columns import Columns
from rich.panel import Panel
from rich.live import Live
from rich.text import Text
from rich.spinner import Spinner, SPINNERS

class WaitingWidget:

    def __init__(self):
        self.live = Live(Spinner("dots", text=Text("Starting", style="green")), refresh_per_second=10)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.live.stop()

    def update_text(self, text):
        self.live.renderable.update(text=Text(text, style="green"))


def make_waiting_widget():

    return WaitingWidget()

def v2():
    # need to be a context manager?
    with make_waiting_widget() as widget:
        for i in range(40):
            if i < 20:
                current_status = "PENDING"
            else:
                current_status = "RUNNING"
            text = f"Waiting job to finish, current status {current_status} (waited for {i}s)"
            widget.update_text(text)
            sleep(0.1)


def v1():
    spinner_name = "dots"
    current_status = "PENDING"
    text = f"Waiting job to finish, current status {current_status}"
    with Live(Spinner(spinner_name, text=Text(text, style="green")), refresh_per_second=10) as live:
        for i in range(40):
            if i < 20:
                current_status = "PENDING"
            else:
                current_status = "RUNNING"
            text = f"Waiting job to finish, current status {current_status} (waited for {i}s)"
            live.renderable.update(text=Text(text, style="green"))
            sleep(0.1)

# v1()
v2()


import dataclasses
import json
import logging
import os
import re
import shutil
import time
from _socket import gaierror
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import NamedTuple, List, Tuple
import traceback

import pandas as pd
import paramiko
from invoke import UnexpectedExit
from paramiko.ssh_exception import AuthenticationException

from slurmpilot.callback import SlurmSchedulerCallback
from slurmpilot.config import Config, load_config
from slurmpilot.jobpath import JobPathLogic
from slurmpilot.remote_command import (
    RemoteCommandExecutionFabrik,
    RemoteExecution,
)
from slurmpilot.util import print_table

logger = logging.getLogger(__name__)
logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)


@dataclass
class JobCreationInfo:
    jobname: str
    entrypoint: str | None = None

    bash_setup_command: str | None = (
        None  # if specified a bash command that gets executed before the main script
    )
    src_dir: str | None = None
    exp_id: str | None = None

    sbatch_arguments: str | None = None  # argument to be passed to sbatch

    # python
    python_binary: str | None = None
    # arguments to be passed to python script, if dictionary then arguments
    # are converted to string with `--key=value` for all key, values of the dictionary
    python_args: str | dict | None = None
    # path existing remotely to be included in PYTHONPATH so that they can be imported in python
    python_paths: list[str] | None = None
    # python libraries existing locally to be sent to the remote and added to the PYTHONPATH
    python_libraries: List[str] | None = None

    # ressources
    cluster: str = None
    partition: str = None
    n_cpus: int = 1  # number of cores
    n_gpus: int = None  # number of gpus per node
    mem: int = None  # memory pool for each core in MB
    max_runtime_minutes: int = 60  # max runtime in minutes
    account: str = None
    env: dict = None

    def __post_init__(self):
        if self.python_args:
            if self.python_binary is None:
                logging.warning(
                    f"Python binary not set but passing `python_args`: {self.python_args}."
                )
        if self.python_binary is not None:
            assert (
                Path(self.entrypoint).suffix == ".py"
            ), f"Must provide a python script ending with .py when using `python_binary` but got {self.entrypoint}."
        if self.src_dir is None:
            self.src_dir = "./"

    def check_path(self):
        assert Path(
            self.src_dir
        ).exists(), f"The src_dir path {self.src_dir} is missing."
        entrypoint_path = Path(self.src_dir) / self.entrypoint
        assert (
            entrypoint_path.exists()
        ), f"The entrypoint could not be found at {entrypoint_path}."

    def sbatch_preamble(self) -> str:
        """
        Spits a preamble like this one valid for sbatch:
        #SBATCH -p {partition}
        #SBATCH --mem {mem}
        ...
        :return:
        """
        res = ""
        sbatch_line = lambda config: f"#SBATCH {config}\n"
        res += sbatch_line(f"--job-name={self.jobname}")
        res += sbatch_line(f"--output=logs/stdout")
        res += sbatch_line(f"--error=logs/stderr")
        if self.n_cpus:
            res += sbatch_line(f"--cpus-per-task={self.n_cpus}")
        if self.partition:
            res += sbatch_line(f"--partition={self.partition}")
        if self.mem:
            res += sbatch_line(f"--mem={self.mem}")
        if self.n_gpus and self.n_gpus > 0:
            res += sbatch_line(f"--gres=gpu:{self.n_gpus}")
        if self.account:
            res += sbatch_line(f"--account={self.account}")
        if self.max_runtime_minutes:
            assert isinstance(
                self.max_runtime_minutes, int
            ), "maxruntime must be an integer expressing the number of minutes"
            res += sbatch_line(f"--time={self.max_runtime_minutes}")
        # res += sbatch_line("--chdir .")
        return res


class JobStatus:
    completed: str = "COMPLETED"
    pending: str = "PENDING"
    failed: str = "FAILED"
    running: str = "RUNNING"
    cancelled: str = "CANCELLED"
    timeout: str = "TIMEOUT"
    out_of_memory: str = "OUT_OF_MEMORY"

    def statuses(self):
        return [self.completed, self.pending, self.failed, self.running, self.cancelled]


@dataclass
class JobMetadata:
    user: str
    date: str
    job_creation_info: JobCreationInfo
    cluster: str

    @property
    def jobname(self):
        return self.job_creation_info.jobname

    def to_json(self) -> str:
        # The methods `to_json` and `from_json` are there because we have nested dataclasses which makes JobMetadata
        # not directly Json serializable
        class EnhancedJSONEncoder(json.JSONEncoder):
            def default(self, o):
                if dataclasses.is_dataclass(o):
                    return dataclasses.asdict(o)
                return super().default(o)

        return json.dumps(self, cls=EnhancedJSONEncoder)

    @classmethod
    def from_json(cls, string) -> "JobMetadata":
        dict_from_string = json.loads(string)
        dict_from_string["job_creation_info"] = JobCreationInfo(
            **dict_from_string["job_creation_info"]
        )
        return JobMetadata(
            **dict_from_string,
        )


class SlurmWrapper:
    def __init__(
        self,
        config: Config | None = None,
        clusters: List[str] | None = None,
        check_connection: bool = True,
        display_loaded_configuration: bool = False,
    ):
        if config is not None:
            self.config = config
        else:
            self.config = load_config()
        if clusters is None:
            clusters = list(self.config.cluster_configs.keys())
        for cluster in clusters:
            assert cluster in self.config.cluster_configs, (
                f"config not found for cluster {cluster}, "
                f"available clusters: {self.config.cluster_configs.keys()}"
            )
        self.clusters = clusters
        self.job_scheduling_callback = SlurmSchedulerCallback()

        if display_loaded_configuration:
            self.job_scheduling_callback.on_config_loaded(self.config)
        self.connections = {}

        # dictionary from cluster name to home dir
        self.home_dir = {}
        for cluster, config in self.config.cluster_configs.items():
            if cluster in clusters:
                self.connections[cluster] = RemoteCommandExecutionFabrik(
                    master=config.host,
                    user=config.user if config.user else os.getenv("USER"),
                    prompt_for_login_password=config.prompt_for_login_password,
                    prompt_for_login_passphrase=config.prompt_for_login_passphrase,
                )
                if check_connection:
                    try:
                        logger.debug(f"Try sending command to {cluster}.")
                        self.job_scheduling_callback.on_establishing_connection(
                            cluster=cluster
                        )
                        home_res = self.connections[cluster].run("echo $HOME")
                        self.home_dir[cluster] = Path(home_res.stdout.strip("\n"))
                    except (gaierror, AuthenticationException) as e:
                        traceback.print_exc()
                        raise ValueError(
                            f"Cannot connect to cluster {cluster}. Verify your ssh access. Error: {str(e)}"
                        )

    def list_clusters(self, cluster: str | None = None) -> List[str]:
        # return a list consisting of the provided cluster if not None or all the clusters if None
        return [cluster] if cluster else self.clusters

    def schedule_job(self, job_info: JobCreationInfo, dryrun: bool = False) -> int:
        # TODO support passing all arguments directly instead of intermediate object
        """
        remote location {slurmpilot_remote}/{jobname} stores:
        * stderr/
        * stdout/
        * slurm_command.sh
        * possible artifacts written by script
        * source dir provided in job_info
        :param dryrun: whether to run a dry run mode which only sends the source remotely but does
        not schedule the job on slurm
        :return: slurm job id if not running in `dry_run` mode
        """
        job_info.check_path()
        cluster_connection = self.connections[job_info.cluster]
        home_dir = self.home_dir[job_info.cluster]
        self.job_scheduling_callback.on_job_scheduled_start(
            cluster=job_info.cluster, jobname=job_info.jobname
        )

        # generate slurm launcher script in slurmpilot dir
        local_job_paths = self._generate_local_folder(job_info)

        # tar and send slurmpilot dir
        remote_job_paths = self.remote_path(
            job_info, root_path=str(home_dir / "slurmpilot")
        )  # TODO clean this
        self.job_scheduling_callback.on_sending_artifact(
            localpath=str(local_job_paths.resolve_path()),
            remotepath=str(remote_job_paths.resolve_path()),
            cluster=job_info.cluster,
            # f"Preparing local folder at {local_job_paths.resolve_path()}"
        )

        cluster_connection.upload_folder(
            local_job_paths.job_path(), remote_job_paths.job_path().parent
        )

        # call sbatch remotely
        if not dryrun:
            jobid = self._call_sbatch_remotely(
                cluster_connection=cluster_connection,
                local_job_paths=local_job_paths,
                remote_job_path=remote_job_paths,
                job_info=job_info,
                sbatch_env=job_info.env if job_info.env else {},
                sbatch_arg=job_info.sbatch_arguments,
            )
            self.job_scheduling_callback.on_job_submitted_to_slurm(
                jobname=job_info.jobname, jobid=jobid
            )
            return jobid
        else:
            return -1

    def stop_job(self, jobname: str):
        metadata = self.job_creation_metadata(jobname)
        jobid = self.jobid_from_jobname(jobname)
        self.connections[metadata.cluster].run(f"scancel {jobid}")

    def wait_completion(
        self,
        jobname: str,
        max_seconds: int = 60,
    ) -> str:
        """
        :param max_seconds: waits for the given number of seconds if defined else waits for status COMPLETED or FAILED
        :return: final status polled
        """
        self.job_scheduling_callback.on_suggest_command_before_wait_completion(
            jobname=jobname
        )
        from rich.live import Live
        from rich.text import Text
        from rich.spinner import Spinner

        starttime = time.time()
        spinner_name = "dots"
        current_status = self.status([jobname])[0]
        wait_interval = 1
        i = 0
        while current_status is None and wait_interval * i < max_seconds:
            time.sleep(wait_interval)
            current_status = self.status([jobname])[0]
            i += 1

        text = f"Waiting job to finish, current status {current_status}"
        with Live(
            Spinner(spinner_name, text=Text(text)),
        ) as live:
            i = 0
            while (
                current_status in [JobStatus.pending, JobStatus.running]
                and wait_interval * i < max_seconds
            ):
                current_status = self.status([jobname])[0]
                text = (
                    f"Waiting job to finish, current status {current_status} (updated every {wait_interval}s, "
                    f"waited for {int(time.time() - starttime)}s)"
                )
                live.renderable.update(text=Text(text))
                time.sleep(
                    wait_interval
                )  # todo exponential backoff to avoid QOS issues
                i += 1
        return current_status

    def _call_sbatch_remotely(
        self,
        job_info: JobCreationInfo,
        cluster_connection: RemoteExecution,
        local_job_paths: JobPathLogic,
        remote_job_path: JobPathLogic,
        sbatch_env: dict,
        sbatch_arg: str | None,
    ) -> int:
        if not sbatch_env:
            sbatch_env = {}
        # TODO design and document SP env that are passed
        sbatch_env["SP_JOBNAME"] = job_info.jobname
        # call sbatch remotely and returns slurm job id if successful
        if sbatch_env:
            # pass environment variable to sbatch in the following form:
            # sbatch --export=REPS=500,X='test' slurm_script.sh
            export_env = "--export=ALL,"
            export_env += ",".join(
                f"{var_name}={var_value}" for var_name, var_value in sbatch_env.items()
            )
        try:
            # TODO make those idempotent,
            #  running `sbatch slurm_script.sh` or `sbatch path_to_script/slurm_script.sh`
            #  should work no matter the directory and variables should be saved for easier reproducibility
            if sbatch_arg is None:
                sbatch_arg = ""
            res = cluster_connection.run(
                f"cd {remote_job_path.job_path()}; mkdir -p logs/; sbatch {sbatch_arg} {export_env} slurm_script.sh",
                env={
                    "SLURMPILOT_PATH": remote_job_path.slurmpilot_path(),
                    "SLURMPILOT_JOBPATH": remote_job_path.resolve_path(),
                },
                retries=3,
            )
        except UnexpectedExit as e:
            raise ValueError(
                "Could not execute sbatch on the remote host, error:" + str(e)
            )
        if res.failed:
            raise ValueError(
                f"Could not submit job, got the following error:\n{res.stderr}"
            )
        elif len(res.stderr) > 0:
            raise ValueError(
                f"Could not submit job, got the following error:\n{res.stderr}"
            )
        else:
            # should be something like 'Submitted batch job 11301013'
            stdout = res.stdout
            slurm_submitted_msg = "Submitted batch job "
            if stdout.startswith(slurm_submitted_msg):
                matches = re.match(slurm_submitted_msg + r"(\d*)", stdout)
                logging.debug(stdout)
                jobid = int(matches.groups()[0])
                # Save the jobid locally so that we can later maps jobname to jobid for slurm query
                # We do not save it remotely as it does not seem necessary but we could if needed
                self._save_jobid(local_job_paths, jobid)
                return jobid
            else:
                raise ValueError(
                    f"Job scheduled without error but could not parse slurm output: {stdout}"
                )

    def remote_path(self, job_info: JobCreationInfo, root_path: str | None = None):
        return JobPathLogic(
            jobname=job_info.jobname,
            entrypoint=job_info.entrypoint,
            src_dir_name=Path(job_info.src_dir).name if job_info.src_dir else None,
            root_path=(
                root_path
                if root_path
                else self.config.remote_slurmpilot_path(job_info.cluster)
            ),
        )

    def _generate_local_folder(self, job_info: JobCreationInfo):
        local_job_paths = JobPathLogic(
            jobname=job_info.jobname,
            entrypoint=job_info.entrypoint,
            src_dir_name=Path(job_info.src_dir).resolve().name,
            root_path=str(self.config.local_slurmpilot_path()),
        )
        assert (
            not local_job_paths.job_path().exists()
        ), f"jobname {job_info.jobname} has already been used, jobnames must be unique please use another one"
        # copy source, generate main slurm script and metadata
        # TODO, option to keep only python files
        shutil.copytree(src=job_info.src_dir, dst=local_job_paths.src_path())
        if job_info.python_libraries:
            for python_library in job_info.python_libraries:
                assert Path(
                    python_library
                ).exists(), (
                    f"Python library specified {python_library} does not exists."
                )
                shutil.copytree(
                    src=python_library,
                    dst=local_job_paths.resolve_path(Path(python_library).name),
                )
        self._generate_main_slurm_script(local_job_paths, job_info)
        self._generate_metadata(local_job_paths, job_info)
        return local_job_paths

    def _generate_main_slurm_script(
        self, local_job_paths: JobPathLogic, job_info: JobCreationInfo
    ):
        with open(local_job_paths.slurm_entrypoint_path(), "w") as f:
            f.write("#!/bin/bash\n")
            f.write(job_info.sbatch_preamble())
            # Add path containing the library to the PYTHONPATH so that they can be imported without requiring
            # the user to add `PYTHONPATH=.` before running scripts, e.g. instead of having to do
            # `PYTHONPATH=. python main.py`, users can simply do `python main.py`
            if job_info.bash_setup_command:
                f.write(job_info.bash_setup_command + "\n")

            # add library to PYTHONPATH
            libraries = [str(local_job_paths.resolve_path())]
            if job_info.python_paths is not None:
                libraries += job_info.python_paths
            if job_info.python_libraries:
                libraries += job_info.python_libraries
            f.write(f'export PYTHONPATH=$PYTHONPATH:{":".join(libraries)}\n')
            if job_info.python_binary is None:
                f.write(f"bash {local_job_paths.entrypoint_path_from_cwd()}\n")
            else:
                python_args = (
                    job_info.python_args if job_info.python_args is not None else ""
                )
                if isinstance(python_args, dict):
                    python_args = " ".join(
                        f"--{key}={value}" for key, value in python_args.items()
                    )
                f.write(
                    f"{job_info.python_binary} {local_job_paths.entrypoint_path_from_cwd()} {python_args}\n"
                )

    def _save_jobid(self, local_job_paths: JobPathLogic, jobid: int):
        metadata = {
            "jobid": jobid,
        }
        with open(local_job_paths.jobid_path(), "w") as f:
            f.write(json.dumps(metadata))

    def log(self, jobname: str, cluster: str | None = None):
        if cluster is None:
            cluster = self.cluster(jobname)
        local_path = JobPathLogic(jobname=jobname)
        self._download_logs(local_path, cluster=cluster, jobname=jobname)
        if local_path.stderr_path().exists():
            with open(local_path.stderr_path()) as f:
                stderr = "".join(f.readlines())
        else:
            stderr = None
        if local_path.stdout_path().exists():
            with open(local_path.stdout_path()) as f:
                stdout = "".join(f.readlines())
        else:
            stdout = None
        return stdout, stderr

    def print_log(self, jobname: str | None = None):
        """
        :param jobname: if not specified uses the last job
        :return: print log of specified job
        """
        if jobname is None:
            jobname = self.latest_job(self.config).jobname
            print(f"No jobname was passed, showing log of latest job {jobname}")
        cluster = self.cluster(jobname)
        stderr, stdout = self.log(jobname=jobname, cluster=cluster)
        if stderr:
            print(f"stderr:\n{stderr}")
        if stdout:
            print(f"stdout:\n{stdout}")

    @staticmethod
    def latest_job(
        config: Config | None = None, pattern: str | None = None
    ) -> JobMetadata:
        """
        :param config:
        :param pattern: if passed, then consider only jobs whose name contains `pattern`
        :return:
        """
        if config is None:
            config = load_config()
        files = list(
            (config.local_slurmpilot_path() / "jobs")
            .expanduser()
            .rglob("metadata.json")
        )
        if pattern is not None:
            # Search for job that matches the patter given
            files = [x for x in files if pattern in str(x)]
        if len(files) > 0:
            latest_file = max(files, key=lambda item: item.stat().st_ctime)
            with open(latest_file, "r") as f:
                return JobMetadata.from_json(f.read())
        raise ValueError(f"No job was found at {config.local_slurmpilot_path()}")

    def cluster(self, jobname: str):
        """
        :param jobname:
        :return: retrieves the cluster where `jobname` was launched by querying local files
        """
        job_metadata = self.job_creation_metadata(jobname=jobname)
        return job_metadata.cluster

    def list_jobs(self, n_jobs: int) -> List[JobMetadata]:
        files = sorted(
            (self.config.local_slurmpilot_path() / "jobs")
            .expanduser()
            .rglob("metadata.json"),
            key=lambda item: item.stat().st_ctime,
            reverse=True,
        )
        jobs = []
        files = list(files)
        files = files[:n_jobs]
        for file in files:
            with open(file, "r") as f:
                try:
                    jobs.append(JobMetadata.from_json(f.read()))
                except json.decoder.JSONDecodeError:
                    pass
        return jobs

    def status(self, jobnames: list[str]) -> list[str | None]:
        if not isinstance(jobnames, list):
            assert isinstance(jobnames, str)
            jobnames = [jobnames]
        jobnames_statuses = {}
        jobid_mapping = {}
        clusters = defaultdict(list)
        # first, we build a dictionary mapping clusters to job informations
        for jobname in jobnames:
            # TODO support having this one missing (
            jobid = self.jobid_from_jobname(jobname)
            jobid_mapping[jobid] = jobname
            job_metadata = self.job_creation_metadata(jobname)
            cluster = job_metadata.cluster
            clusters[cluster].append((jobid, job_metadata))

        # second we call sacct on each clusters with the corresponding jobs
        for cluster in clusters.keys():
            try:
                job_clusters = clusters[cluster]
                # filter jobs with missing jobid
                query_jobids = [
                    str(jobid)
                    for (jobid, job_metadata) in job_clusters
                    if jobid is not None
                ]
                sacct_format = "JobID,Elapsed,start,State"

                # call sacct...
                res = self.connections[cluster].run(
                    f'sacct --format="{sacct_format}" -X -p --jobs={",".join(query_jobids)}'
                )

                # ...and parse output
                lines = res.stdout.split("\n")
                keys = lines[0].split("|")[:-1]
                for line in lines[1:]:
                    if line:
                        try:
                            status_dict = dict(zip(keys, line.split("|")[:-1]))
                            # TODO elapsed time is probably useful too
                            jobnames_statuses[
                                jobid_mapping[int(status_dict["JobID"])]
                            ] = status_dict.get("State", "missing")
                        except Exception as e:
                            print(line, str(e))
                            continue
            # TODO abstract it
            except paramiko.ssh_exception.AuthenticationException as e:
                logging.warning(
                    f"Could not connect to cluster {cluster}, make sure your ssh connection works."
                )
                continue
        return [jobnames_statuses.get(jobname) for jobname in jobnames]

    def print_jobs(
        self, n_jobs: int = 10, max_colwidth: int = 50, status_verbose: bool = True
    ):
        rows = []
        jobs = self.list_jobs(n_jobs)
        statuses = self.status(jobnames=[job.jobname for job in jobs])
        for job, status in zip(jobs, statuses):
            # TODO status when missing jobid.json => error at launching
            if status is None:
                status_symbol = "Unknown slurm status 🏝"
            else:
                status_mapping = {
                    JobStatus.out_of_memory: "OOM 🤯",
                    JobStatus.failed: "Slurm job failed ❌",
                    JobStatus.pending: "Pending ⏳",
                    JobStatus.running: "️Running 🏃",
                    JobStatus.completed: "Completed ✅",
                    JobStatus.cancelled: "Canceled ⚠️",
                }
                if "CANCELLED" in status:
                    status_symbol = "Cancelled ⚠️"
                else:
                    status_symbol = status_mapping.get(
                        status, f"Unknown state: {status}"
                    )

            rows.append(
                {
                    "job": Path(job.jobname).name,
                    "date": pd.to_datetime(job.date).strftime("%d/%m/%y-%H:%M"),
                    "cluster": job.job_creation_info.cluster,
                    "status": (
                        f"{status_symbol} "
                        if status_verbose
                        else f"{status_symbol[-1:]} "
                    ),
                    "full jobname": job.jobname,
                }
            )
        df = pd.DataFrame(rows)
        print(
            df.to_string(
                index=False,
                max_colwidth=max_colwidth,
            )
        )

    def download_job(self, jobname: str | None = None):
        if jobname is None:
            jobname = self.latest_job(self.config).jobname
        cluster = self.cluster(jobname)
        local_path = JobPathLogic(jobname=jobname)
        remote_path = JobPathLogic.from_jobname(
            jobname=jobname,
            root_path=self.config.remote_slurmpilot_path(cluster),
        )
        self.connections[cluster].download_folder(
            remote_path.resolve_path(),
            local_path.resolve_path(),
        )

    def _download_logs(self, local_path, cluster: str, jobname: str):
        # 1) download log from remote file
        # 2) show in console
        # we could also consider streaming
        remote_path = JobPathLogic.from_jobname(
            jobname=jobname,
            root_path=self.config.remote_slurmpilot_path(cluster),
        )
        try:
            self.connections[cluster].download_file(
                remote_path.stderr_path(), local_path.stderr_path()
            )
            self.connections[cluster].download_file(
                remote_path.stdout_path(), local_path.stdout_path()
            )
        except FileNotFoundError:
            return ""

    def jobid_from_jobname(self, jobname: str) -> int | None:
        # several options are possible:
        # 1) store info in jobname folder
        # 2) runs sacct -X --format=jobname%30,jobid remotely and parse output
        # 3) store mapping locally in DB
        # We choose here the first one because it avoids a remote/call parsing or to setup a DB

        # Option 2)
        # res = self.connection.run(f"sacct -X --format=jobname%30,jobid  | grep {jobname} | tail -n 1")
        # # "example-j+ 11301527" -> 11301527
        # return int(res.stdout.split(" ")[-1])
        local_path = JobPathLogic(jobname=jobname)
        if local_path.jobid_path().exists():
            with open(local_path.jobid_path(), "r") as f:
                return json.load(f)["jobid"]
        else:
            return None

    @staticmethod
    def job_creation_metadata(jobname: str) -> JobMetadata:
        local_path = JobPathLogic(jobname=jobname)
        with open(local_path.metadata_path(), "r") as f:
            return JobMetadata.from_json(f.read())

    def _generate_metadata(self, local_path: JobPathLogic, job_info: JobCreationInfo):
        metadata = JobMetadata(
            user=os.getenv("USER"),
            date=str(datetime.now()),
            job_creation_info=job_info,
            cluster=job_info.cluster,
        )
        with open(local_path.metadata_path(), "w") as f:
            f.write(metadata.to_json())

    def format_string_jobname(self, message: str, jobname: str) -> str:
        return self.job_scheduling_callback.format_string_jobname(message, jobname)



import sys
import time
def print_statusline(msg: str):
    last_msg_length = len(getattr(print_statusline, 'last_msg', ''))
    print(' ' * last_msg_length, end='\r')
    print(msg, end='\r')
    # sys.stdout.flush()  # Some say they needed this, I didn't.
    setattr(print_statusline, 'last_msg', msg)

for msg in ["Initializing...", "Initialization successful!"]:
    print_statusline(msg)
    time.sleep(1)


import pprint
import time

from slurmpilot.config import Config


class SlurmSchedulerCallbackInterface:
    def on_job_scheduled_start(self, cluster: str, jobname: str):
        raise NotImplementedError()

    def on_established_connection(self, cluster: str):
        raise NotImplementedError()

    def on_sending_artifact(self, localpath: str, remotepath: str, cluster: str):
        raise NotImplementedError()

    def on_job_submitted_to_slurm(self, jobid: int, jobname: str):
        raise NotImplementedError()


class SlurmSchedulerCallback(SlurmSchedulerCallbackInterface):
    def __init__(self):
        self.format_cluster = "1;30;34"
        self.format_highlight = "1;30;34"
        self.format_jobname = "1;31;38"

    def format(self, s, format_pattern=None):
        if not format_pattern:
            format_pattern = self.format_cluster
        return f"\x1b[{format_pattern}m{s}\x1b[0m"

    def format_string_jobname(self, message: str, jobname: str) -> str:
        return f"{message} {self.format(jobname, self.format_jobname)}."

    def on_job_scheduled_start(self, cluster: str, jobname: str):
        print(
            f"Starting job {self.format(jobname, self.format_jobname)} on {self.format(cluster, self.format_cluster)}."
        )

    def on_establishing_connection(self, cluster: str):
        print(
            f"Establishing ssh connection with {self.format(cluster, self.format_cluster)}"
        )

    def on_sending_artifact(self, localpath: str, remotepath: str, cluster: str):
        print(f"Sending job data from {localpath} to {cluster}:{remotepath}")

    def on_job_submitted_to_slurm(self, jobid: int, jobname: str):
        print(
            f"Job submitted to Slurm with the following id {self.format(jobid, self.format_jobname)} saving the jobid locally."
        )

    def on_suggest_command_before_wait_completion(self, jobname: str):
        commands = [
            (
                "show the log of your job",
                f"slurmpilot --log {jobname}",
            ),
            (
                "sync the artifact of your job",
                f"slurmpilot --sync {jobname}",
            ),
            (
                "show the status of your job",
                f"slurmpilot --status {jobname}",
            ),
            (
                "stop your job",
                f"slurmpilot --stop {jobname}",
            ),
        ]
        commands_strings = []
        for description, command in commands:
            commands_strings.append(
                f"* {description}: "
                + self.format(f"`{command}`", self.format_highlight)
            )
        cmds = "\n".join(commands_strings)
        print(f"You can use the following commands in a terminal:\n" + cmds)

    def on_waiting_completion(self, jobname: str, status: str, n_seconds_wait: int):
        # TODO dependency inversion to support rich
        print(
            f"{self.format(jobname, self.format_jobname)} status {self.format(status, self.format_highlight)}, waiting {n_seconds_wait}s"
        )

    def on_config_loaded(self, config: Config):
        print(f"Cluster configurations loaded:")
        for cluster, cluster_config in config.cluster_configs.items():
            print(f"{self.format(cluster, self.format_cluster)}: {cluster_config}")


if __name__ == "__main__":

    def print_format_table():
        """
        prints table of formatted text format options
        """
        for style in range(8):
            for fg in range(30, 38):
                s1 = ""
                for bg in range(30, 48):
                    format = ";".join([str(style), str(fg), str(bg)])
                    s1 += "\x1b[%sm %s \x1b[0m" % (format, format)
                print(s1)
            print("\n")

    print_format_table()

    print("\x1b[0;31;40m" + "Success!" + "\x1b[0m" + "yop")

    cb = SlurmSchedulerCallback()
    cluster = "bigcluster"
    jobname = "smalljob"
    cb.on_job_scheduled_start(cluster=cluster, jobname=jobname)
    cb.on_establishing_connection(cluster=cluster)
    cb.on_sending_artifact(cluster=cluster, localpath="foo/", remotepath="foo2/")
    cb.on_job_submitted_to_slurm(jobname=jobname, jobid=12)
    cb.on_suggest_command_before_wait_completion(jobname=jobname)
    # cb.on_config_loaded()
    for _ in range(4):
        cb.on_waiting_completion(jobname=jobname, status="PENDING", n_seconds_wait=1)
        time.sleep(0.2)



from time import sleep

from rich.columns import Columns
from rich.panel import Panel
from rich.live import Live
from rich.text import Text
from rich.spinner import Spinner, SPINNERS

all_spinners = Columns(
    [
        Spinner(spinner_name, text=Text(repr(spinner_name), style="green"))
        for spinner_name in sorted(SPINNERS)
    ],
    column_first=True,
    expand=True,
)

with Live(
    Panel(all_spinners, title="Spinners", border_style="blue"),
    refresh_per_second=20,
) as live:
    while True:
        sleep(0.2)



import logging
import socket
import subprocess
import tempfile
import time
from dataclasses import dataclass
from getpass import getpass
from pathlib import Path
import tarfile

import paramiko

logger = logging.getLogger(__name__)


@dataclass
class CommandResult:
    command: str
    failed: bool
    stderr: str
    stdout: str
    return_code: int


class RemoteExecution:
    def __init__(self, master: str, user: str | None = None, proxy: str | None = None):
        """
        Wraps functionality to run remote command and upload/download files and folders to a remote host possibly
        connected through a proxy.
        Note: We wrap fabric instead of exposing it (which is a wrapper itself) in order to be able to support
        multiple choices in the future possibly including one which is dependency free.
        :param master:
        :param proxy:
        """
        self.master = master
        self.proxy = proxy
        self.user = user

    def run(
        self, command: str, pty: bool = False, env: dict | None = None, retries: int = 0
    ) -> CommandResult:
        raise NotImplementedError()

    def upload_file(self, local_path: Path, remote_path: Path = Path("/")):
        raise NotImplementedError()

    def upload_folder(self, local_path: Path, remote_path: Path = Path("/")):
        raise NotImplementedError()

    def download_file(self, remote_path: Path, local_path: Path):
        raise NotImplementedError()

    def download_folder(self, remote_path: Path, local_path: Path):
        raise NotImplementedError()


class LocalCommandExecution(RemoteExecution):
    def __init__(
        self,
        master: str,
        user: str | None = None,
        proxy: str | None = None,
        folder: str | None = None,
    ):
        super().__init__(master=master, proxy=proxy, user=user)
        self.folder = folder

    def run(
        self,
        command: str,
        pty: bool = False,
        env: dict | None = None,
        retries: int = 0,
        log_error: bool = True,
    ) -> CommandResult:
        # TODO evaluate command with subprocess like in syne tune
        raise NotImplementedError()

    def upload_file(self, local_path: Path, remote_path: Path = Path("/")):
        # TODO copy file locally
        raise NotImplementedError()

    def upload_folder(self, local_path: Path, remote_path: Path = Path("/")):
        # TODO copy folder locally with shutil.copytree
        raise NotImplementedError()

    def download_file(self, remote_path: Path, local_path: Path):
        # TODO copy file locally
        raise NotImplementedError()

    def download_folder(self, remote_path: Path, local_path: Path):
        # TODO copy file locally
        raise NotImplementedError()


class RemoteCommandExecutionFabrik(RemoteExecution):
    # TODO we could create a dependency free version with `getstatusoutput` that calls ssh command
    def __init__(
        self,
        master: str,
        user: str | None = None,
        proxy: str | None = None,
        prompt_for_login_password: bool = False,
        prompt_for_login_passphrase: bool = False,
    ):
        from fabric import Connection

        super().__init__(master=master, proxy=proxy, user=user)
        logging.getLogger("fabric").setLevel(logging.WARNING)
        logging.getLogger("paramiko").setLevel(logging.WARNING)
        if prompt_for_login_passphrase or prompt_for_login_password:
            connect_kwargs = {}
            if prompt_for_login_password:
                prompt = "Enter login password for SSH auth: "
                connect_kwargs["password"] = getpass(prompt)
            if prompt_for_login_passphrase:
                prompt = "Enter passphrase for unlocking SSH keys: "
                connect_kwargs["passphrase"] = getpass(prompt)
        else:
            connect_kwargs = {}
        self.connection = Connection(
            self.master,
            user=user,
            gateway=None if not proxy else Connection(proxy),
            connect_kwargs=connect_kwargs,
        )
        # follow the same procedure as in fabric main to set the connect_kwargs which feels awkward,
        # we can consider using another library or direct use of ssh through multiprocessing.
        # https://github.com/fabric/fabric/blob/main/fabric/main.py#L151
        self.connection.config._overrides["connect_kwargs"] = connect_kwargs
        # Since we gave merge=False above, we must do it ourselves here. (Also
        # allows us to 'compile' our overrides manipulation.)
        self.connection.config.merge()

    def run(
        self,
        command: str,
        pty: bool = False,
        env: dict | None = None,
        retries: int = 0,
        log_error: bool = True,
    ) -> CommandResult:
        success = False
        num_trial = 1 + retries
        while not success and num_trial > 0:
            try:
                fabric_result = self.connection.run(
                    command=command,
                    hide=True,
                    pty=pty,
                    env=env,
                )
                success = not fabric_result.failed
                # TODO show error when failed
                if fabric_result.failed:
                    if log_error:
                        logging.debug(
                            f"Command {command} failed\n{fabric_result.stderr}"
                        )
            except paramiko.ssh_exception.ChannelException as e:
                if log_error:
                    logging.debug(
                        f"Command {command} failed because of connection issue {str(e)}"
                    )
            except socket.gaierror as e:
                logging.error(str(e))
                raise ValueError(
                    f"Could not connect to hostname {self.master}, check your ssh connection."
                )
            if not success:
                time.sleep(1)
                num_trial -= 1
        if success:
            return CommandResult(
                command=command,
                failed=fabric_result.failed,
                return_code=fabric_result.return_code,
                stderr=fabric_result.stderr,
                stdout=fabric_result.stdout,
            )
        else:
            raise ValueError(
                f"Command {command} did not succeed after {retries} trial, consider increasing `retries` argument."
            )

    def upload_file(self, local_path: Path, remote_path: Path = Path("/")):
        # TODO consider using rsync which is supported in Fabric,
        #  potential downside: it may not be installed on remote node...
        # fabric.contrib.project.upload_project seems also a viable option
        local_path = Path(local_path)
        assert local_path.is_file()
        self.run(f"mkdir -p {remote_path}")
        self.connection.put(
            local=str(local_path),
            remote=str(remote_path),
        )

    def upload_folder(self, local_path: Path, remote_path: Path = Path("/")):
        local_path = Path(local_path)
        assert local_path.is_dir()
        # tar before sending and untar remotely
        # Note: we could also use rsync
        with tempfile.TemporaryDirectory() as tmpdirname:
            tarpath = self._tar(local_path, tgt=Path(tmpdirname) / local_path.name)
            logger.info(f"sending {tarpath} remotely to {self.connection.host}")
            self.run(f"mkdir -p {str(remote_path)}")
            self.connection.put(
                local=str(tarpath),
                remote=str(remote_path),
            )
            logger.debug(f"extracting remotely")
            command = f"cd {remote_path}; tar -xvf {tarpath.name}; rm {tarpath.name}"
            self.run(command)

    def _tar(self, path_to_archive: Path, tgt: Path):
        assert path_to_archive.is_dir()
        tarpath = tgt.with_suffix(".tar.gz")
        logger.info(f"Compressing {path_to_archive} into {tgt}")
        with tarfile.open(tarpath, "w:gz") as tar:
            tar.add(path_to_archive, arcname=path_to_archive.stem)
        return tarpath

    def download_file(self, remote_path: Path, local_path: Path):
        self.connection.get(
            remote=str(remote_path),
            local=str(local_path),
        )

    def download_folder(self, remote_path: Path, local_path: Path):
        """
        :param remote_path:
        :param local_path:
        :return:
        """
        # Note, we could also tar the whole thing like we do to send, the reason we pick rsync is that often
        # some files will only be present and rsync allows to not copy those based on hashes
        logger.info(f"Running rsync from {remote_path} to {local_path}")
        command = (
            f"rsync -aPvz {self.user}@{self.master}:{remote_path} {local_path.parent}"
        )
        subprocess.run(command.split(" "), check=True)


if __name__ == "__main__":
    connection = RemoteCommandExecutionFabrik(
        master="YOURCLUSTER",
        user="YOURUSER",
        prompt_for_login_password=True,
        prompt_for_passphrase=False,
    )

    print(connection.run("ls"))



import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import NamedTuple, Dict, List, Tuple

import yaml

logger = logging.getLogger(__name__)
logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
config_path = Path("~/slurmpilot/config").expanduser()


class GeneralConfig(NamedTuple):
    # General configurations containing defaults

    # default path where slurmpilot job files are generated
    local_path: str

    # default path where slurmpilot job files are generated on the remote machine, Note: "~" cannot be used
    remote_path: str

    # default cluster to be used, must have a file `config/clusters/{default_cluster}.yaml` associated
    default_cluster: str | None = None


@dataclass
class ClusterConfig:
    # Configuration for a cluster, can override default in GeneralConfig
    host: str
    remote_path: str = "slurmpilot/"
    account: str | None = None
    user: str | None = None
    default_partition: str | None = None
    prompt_for_login_passphrase: bool = False  # whether to prompt for ssh passphrase
    prompt_for_login_password: bool = False  # whether to prompt for ssh password


class Config:
    def __init__(
        self,
        general_config: GeneralConfig | None = None,
        cluster_configs: Dict[str, ClusterConfig] | None = None,
    ):
        self.general_config = general_config
        self.cluster_configs = cluster_configs if cluster_configs is not None else {}

    @classmethod
    def load_from_path(cls, path=config_path, clusters: List[str] | None = None):
        general_config = load_general_config(path / "general.yaml")
        cluster_configs = {
            cluster_config_path.stem: load_cluster_config(cluster_config_path)
            for cluster_config_path in (path / "clusters").rglob("*.yaml")
            if clusters is None or cluster_config_path.stem in clusters
        }
        return cls(general_config, cluster_configs)

    def save_to_path(self, path=config_path, clusters: List[str] | None = None):
        path.mkdir(parents=True, exist_ok=True)
        (path / "clusters").mkdir(parents=True, exist_ok=True)
        if self.general_config:
            with open(path / "general.yaml", "w") as f:
                yaml.dump(self.general_config._asdict(), f)
        if clusters is None:
            clusters = self.cluster_configs.keys()
        for cluster in clusters:
            with open(path / "clusters" / f"{cluster}.yaml", "w") as f:
                dict_without_none = {
                    k: v for k, v in self.cluster_configs[cluster].__dict__.items() if v
                }
                yaml.dump(dict_without_none, f)

    def __str__(self):
        res = {
            "general_config": self.general_config,
            "cluster_configs": "\n".join(
                [f"{k}: {v}" for k, v in self.cluster_configs.items()]
            ),
        }
        return json.dumps(res, indent=4)

    def local_slurmpilot_path(self) -> Path:
        return Path(self.general_config.local_path).expanduser()

    def remote_slurmpilot_path(self, cluster: str) -> Path:
        remote_path = self.cluster_configs[cluster].remote_path
        if remote_path is None:
            remote_path = self.general_config.remote_path
        return Path(remote_path)


def load_yaml(path: Path) -> dict:
    with open(path, "r") as stream:
        return yaml.safe_load(stream)


def load_cluster_config(path: Path) -> ClusterConfig:
    if path.exists():
        return ClusterConfig(**load_yaml(path))
    else:
        return None


def load_general_config(path: Path) -> GeneralConfig:
    if path.exists():
        args = load_yaml(path)
        args["local_path"] = str(Path(args["local_path"]).expanduser())
        return GeneralConfig(**load_yaml(path))
    else:
        return None


def load_config(user_path: Path | None = None) -> Config:
    """
    :param user_path:
    :return: loads configuration by default from ~/slurmpilot/config unless `user_path` is specified.
    """
    if user_path is None:
        user_path = Path("~/slurmpilot/config").expanduser()
    user_config = Config.load_from_path(user_path)

    general_config = user_config.general_config
    if general_config is None:
        general_config = GeneralConfig(
            local_path="~/slurmpilot", remote_path="slurmpilot/"
        )

    logger.info(
        f"Loaded cluster configurations {list(user_config.cluster_configs.keys())}."
    )
    return Config(
        general_config=general_config, cluster_configs=user_config.cluster_configs
    )


def default_cluster_and_partition(user_path: Path | None = None) -> Tuple[str, str]:
    """
    :param user_path:
    :return: default cluster and partition. The values should be specified in "~/slurmpilot/general.yaml" for
    `default_cluster` and the default partition should be specified in the cluster config with
    "~/slurmpilot/configs/{default_cluster}.yaml". Alternatively, one can also specify the default cluster with the
    environment variable "SP_DEFAULT_CLUSTER".
    """
    # We have a couple of options, we could
    # 1) load the configuration and take the first cluster (assuming it has a default partition field)
    # 2) load an environment variable DEFAULT_CLUSTER
    config = load_config(user_path=user_path)

    if "SP_DEFAULT_CLUSTER" in os.environ:
        cluster = os.getenv("SP_DEFAULT_CLUSTER")
        assert cluster in config.cluster_configs
        # TODO add partition to cluster config
        partition = config.cluster_configs[cluster].default_partition
    else:
        cluster = config.general_config.default_cluster
        assert cluster is not None, (
            "To be able to use a default cluster, you need to set the environment variable $SP_DEFAULT_CLUSTER to the "
            "desired cluster or alternatively to add `default_cluster` to general.yaml"
        )
        partition = config.cluster_configs[cluster].default_partition

    assert (
        partition is not None
    ), f"Cannot use default cluster {cluster} without default partition, provide it in {cluster}.yaml"
    return cluster, partition



import argparse
import os

import yaml

from slurmpilot.config import load_config, ClusterConfig, config_path
from slurmpilot.slurm_wrapper import SlurmWrapper


def jobname_from_cli_args_or_take_latest(args):
    if args.job:
        try:
            return SlurmWrapper.job_creation_metadata(args.job)
        except FileNotFoundError as e:
            print(
                f"Jobname passed not found, searching for a jobname which contains {args.job}."
            )
            job_metadata = SlurmWrapper.latest_job(pattern=args.job)
            return job_metadata

    else:
        job_metadata = SlurmWrapper.latest_job()
        print(f"No jobs specified, retrieved the latest one: {job_metadata.jobname}")
        return job_metadata


def main():
    parser = argparse.ArgumentParser(
        prog="Slurmpilot CLI",
        # description='What the program does',
        # epilog='Text at the bottom of help'
    )
    # TODO list all local jobs and all remote jobs
    parser.add_argument(
        "job", help="Jobname to be queried, leave empty to use the last one", nargs="?"
    )
    parser.add_argument(
        "--log",
        help="Show log for the specified job",
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--stop", help="Stop the specified job", action=argparse.BooleanOptionalAction
    )
    parser.add_argument(
        "--download",
        help="Download data for the specified job",
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--status",
        help="Show status for the specified job",
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--list_jobs", help="List n latest jobs", required=False, type=int, default=5
    )

    parser.add_argument(
        "--test-ssh-connections",
        help="Test ssh connections",
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        "--add-cluster", help="Add a new cluster", action=argparse.BooleanOptionalAction
    )

    args = parser.parse_args()
    is_command_requiring_jobname = any(
        [args.log, args.download, args.status, args.stop]
    )
    if (
        not is_command_requiring_jobname
        and not args.test_ssh_connections
        and not args.add_cluster
        and not args.list_jobs
    ):
        raise ValueError(
            "No action specifed, run slurmpilot YOURJOB --log to display the log, or use --status for other actions."
        )

    if is_command_requiring_jobname:
        job = jobname_from_cli_args_or_take_latest(args)
        slurm = SlurmWrapper(clusters=[job.cluster])
        # TODO use match
        if args.log:

            print(slurm.format_string_jobname("Displaying log for job", job.jobname))
            slurm.print_log(jobname=job.jobname)
        if args.download:
            print(slurm.format_string_jobname("Downloading job", job.jobname))
            slurm.download_job(jobname=job.jobname)
        if args.status:
            print(slurm.format_string_jobname("Displaying status for job", job.jobname))
            print(slurm.status(jobnames=[job.jobname])[0])
        if args.stop:
            print(slurm.format_string_jobname("Stopping job", job.jobname))
            slurm.stop_job(jobname=job.jobname)
    else:
        if args.test_ssh_connections:
            slurm = SlurmWrapper(config=load_config())
            print(
                f"Could load successfully ssh connections from {list(slurm.connections.keys())} clusters."
            )
            # TODO print a table with ✅❌ symbols
        elif args.add_cluster:
            cluster = input("Name of the cluster:")
            hostname = input("Hostname:")
            username = input(f'Username (optional, default to {os.getenv("USER")}):')
            account = input("Default account (optional):")
            default_partition = input("Default partition (optional):")
            remote_path = input(f"Remote path (default to slurmpilot/):")

            prompt_for_login_password = (
                input(
                    f"Prompt for a login password for ssh if you set a key password? (y/n) optional and deactivated by default:"
                )
                .lower()
                .strip()
                == "y"
            )
            prompt_for_login_passphrase = (
                input(
                    f"Prompt for a login passphrase for ssh if you set a key password? (y/n) optional and deactivated by default:"
                )
                .lower()
                .strip()
                == "y"
            )

            config_cluster_path = config_path / "clusters" / f"{cluster}.yaml"
            print(f"Saving cluster configuration {cluster} to {config_cluster_path}")
            config = ClusterConfig(
                host=hostname,
                remote_path=remote_path,
                account=account,
                user=username,
                default_partition=default_partition,
                prompt_for_login_passphrase=prompt_for_login_passphrase,
                prompt_for_login_password=prompt_for_login_password,
            )
            print(f"Going to save {config} to {config_cluster_path}")

            with open(config_path / "clusters" / f"{cluster}.yaml", "w") as f:
                dict_without_none = {k: v for k, v in config.__dict__.items() if v}
                yaml.dump(dict_without_none, f)

            try:
                print(f"Testing configuration {cluster} with ssh.")
                SlurmWrapper(config=load_config(), clusters=[cluster])
                print(f"✅ Could connect to ssh, cluster successfully installed.")
            except ValueError as e:
                print(
                    f"⚠️ Could not connect to ssh, you can remove the file {config_cluster_path} or rerun this script"
                )
                # config_cluster_path.unlink()
        elif args.list_jobs:
            n_jobs = args.list_jobs
            SlurmWrapper(check_connection=False, config=load_config()).print_jobs(
                n_jobs=n_jobs
            )


if __name__ == "__main__":
    main()



import random
import string
import time
from contextlib import contextmanager
from time import perf_counter
from typing import Tuple


@contextmanager
def catchtime(name: str, logger=None) -> float:
    start = perf_counter()
    print_fun = print if logger is None else logger.info
    try:
        print_fun(f"start: {name}")
        yield lambda: perf_counter() - start
    finally:
        print_fun(f"Time for {name}: {perf_counter() - start:.4f} secs")


def unify(name: str, method: str = "date") -> str:
    """
    :param name:
    :param method:
    :return: a name likely to be unique obtained by adding a random suffix or a time-stamp depending on the method
    """
    assert method in ["ascii", "coolname", "date"]
    if method == "ascii":
        suffix = "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(5)
        )
    elif method == "coolname":
        from coolname import generate_slug

        suffix = generate_slug()
    else:
        suffix = time.strftime("%Y-%m-%d-%H-%M-%S")
    return name + "-" + suffix


def print_table(rows):
    if len(rows) > 0:
        import pandas as pd

        print(pd.DataFrame(rows).set_index("JobName").to_string())



"""
A simulation of Rich console logging.
"""

import time
from rich.console import Console
from rich.style import Style
from rich.theme import Theme
from rich.highlighter import RegexHighlighter


class RequestHighlighter(RegexHighlighter):
    base_style = "req."
    highlights = [
        r"^(?P<protocol>\w+) (?P<method>\w+) (?P<path>\S+) (?P<result>\w+) (?P<stats>\[.+\])$",
        r"\/(?P<filename>\w+\..{3,4})",
    ]


theme = Theme(
    {
        "req.protocol": Style.parse("dim bold green"),
        "req.method": Style.parse("bold cyan"),
        "req.path": Style.parse("magenta"),
        "req.filename": Style.parse("bright_magenta"),
        "req.result": Style.parse("yellow"),
        "req.stats": Style.parse("dim"),
    }
)
console = Console(theme=theme)

console.log("Server starting...")
console.log("Serving on http://127.0.0.1:8000")

time.sleep(1)

request_highlighter = RequestHighlighter()

console.log(
    request_highlighter("HTTP GET /foo/bar/baz/egg.html 200 [0.57, 127.0.0.1:59076]"),
)

console.log(
    request_highlighter(
        "HTTP GET /foo/bar/baz/background.jpg 200 [0.57, 127.0.0.1:59076]"
    ),
)


time.sleep(1)


def locals():
    foo = (1, 2, 3)
    movies = ["Deadpool", "Rise of the Skywalker"]
    console = Console()

    console.log(
        "[b]JSON[/b] RPC [i]batch[/i]",
        [
            {"jsonrpc": "2.0", "method": "sum", "params": [1, 2, 4], "id": "1"},
            {"jsonrpc": "2.0", "method": "notify_hello", "params": [7]},
            {"jsonrpc": "2.0", "method": "subtract", "params": [42, 23], "id": "2"},
            {"foo": "boo"},
            {
                "jsonrpc": "2.0",
                "method": "foo.get",
                "params": {"name": "myself", "enable": False, "grommits": None},
                "id": "5",
            },
            {"jsonrpc": "2.0", "method": "get_data", "id": "9"},
        ],
        log_locals=True,
    )


locals()
